MODULE mask_mod
  USE par_kind, ONLY: wp
  IMPLICIT none
  PRIVATE

  ! To hold the wet/dry mask for the domain. Obtained from the bathymetry
  ! read from file.
  INTEGER, ALLOCATABLE, DIMENSION(:,:) :: imask

  LOGICAL :: ln_zco, ln_zps, ln_sco        ! Vert. coordinate options

   !* Namelist namzgr_sco *
   REAL(wp) ::  rn_sbot_min = 300._wp    ! minimum depth of s-bottom surface (>0) (m)
   REAL(wp) ::  rn_sbot_max = 5250._wp    ! maximum depth of s-bottom surface (= ocean depth) (>0) (m)
   REAL(wp) ::  rn_theta    =    6.00_wp  ! surface control parameter (0<=rn_theta<=20)
   REAL(wp) ::  rn_thetb    =    0.75_wp  ! bottom control parameter  (0<=rn_thetb<= 1)
   REAL(wp) ::  rn_rmax     =    0.15_wp  ! maximum cut-off r-value allowed (0<rn_rmax<1)
   LOGICAL  ::  ln_s_sigma  = .false.     ! use hybrid s-sigma -coordinate & stretching function fssig1 (ln_sco=T)
   REAL(wp) ::  rn_bb       =    0.80_wp  ! stretching parameter for song and haidvogel stretching
   !                                       ! ( rn_bb=0; top only, rn_bb =1; top and bottom)
   REAL(wp) ::  rn_hc       =  150._wp    ! Critical depth for s-sigma coordinates

  PUBLIC imask
  PUBLIC generate_mask

CONTAINS

  SUBROUTINE read_namelist_options(ierr)
    !!---------------------------------------------------------------------
    ! Read the z-coordinate options from the namelist file
    !!---------------------------------------------------------------------
    IMPLICIT none
    INTEGER, INTENT(out) :: ierr
    ! Locals
    INTEGER, PARAMETER :: numnam = 31
    ! From OPA_SRC/DOM/domzgr
    NAMELIST/namzgr/ ln_zco, ln_zps, ln_sco
    NAMELIST/namzgr_sco/ rn_sbot_max, rn_sbot_min, rn_theta, rn_thetb, &
                         rn_rmax, ln_s_sigma, rn_bb, rn_hc

    OPEN( UNIT=numnam, FILE='namelist', STATUS='OLD', FORM='FORMATTED', &
         ACCESS='SEQUENTIAL',IOSTAT=ierr )
    IF(ierr /= 0)THEN
       WRITE(*,*) 'read_namelist_options: ERROR: failed to open file ''namelist'''
    END IF
      
    REWIND(numnam)
    READ  (numnam, namzgr)

    IF(ln_sco)THEN
       REWIND(numnam)
       READ(numnam, namzgr_sco)
    END IF

    CLOSE(numnam)

  END SUBROUTINE read_namelist_options


  SUBROUTINE generate_mask(regular, ierr)
    !!---------------------------------------------------------------------
    !!---------------------------------------------------------------------
    USE netcdf
    USE partition_mod, ONLY: LAND, jpiglo, jpjglo
    IMPLICIT none
    LOGICAL, INTENT(In) :: regular
    INTEGER, INTENT(out) :: ierr
    ! Local vars
    INTEGER :: if90id           ! nf90 identifier of the opened file
    INTEGER :: ivarid           ! netcdf variable ID
    INTEGER :: i_dimid_lat, i_dimid_lon ! IDs of dimensions in netcdf file
    INTEGER :: ilat, ilon               ! Current values of those dimensions
    REAL(KIND(1.0D0)), ALLOCATABLE, DIMENSION(:,:) :: bathy
    !!---------------------------------------------------------------------  

    ! Initialise error flag
    ierr = 0

    ! We read the NEMO namelist file to establish what z-coordinate
    ! options are being used. If ln_sco is .TRUE. then we read
    ! the parameters used in smoothing the bathymetry.
    CALL read_namelist_options(ierr)
    IF( ierr /= 0)THEN
       WRITE(*,*) 'Failed to read z-coord options from namelist file'
       RETURN
    END IF

    IF(.NOT. regular)THEN
       ! ============================
       ! Generate a global mask from the model bathymetry
       ! ============================

       ! Read the bathymetry from file
       ! Open the file
       ierr = NF90_OPEN( 'bathy_meter.nc', NF90_NOWRITE  , if90id )
       IF(ierr /= nf90_noerr)THEN
          WRITE(*,*) 'generate_mask: Failed to open netcdf file bathy_meter.nc'
          ierr = 1
          RETURN
       END IF

       IF(jpiglo < 1 .OR. jpjglo < 1)THEN

          ierr = NF90_INQ_DIMID(if90id, 'lat', i_dimid_lat )
          IF(ierr /= nf90_noerr)THEN
             WRITE(*,*) 'generate_mask: Failed to read ID of lat dimension from netcdf file bathy_meter.nc'
             ierr = 1
             RETURN
          END IF

          ierr = NF90_INQ_DIMID(if90id, 'lon', i_dimid_lon )
          IF(ierr /= nf90_noerr)THEN
             WRITE(*,*) 'generate_mask: Failed to read ID of lon dimension from netcdf file bathy_meter.nc'
             ierr = 1
             RETURN
          END IF

          ierr = NF90_INQUIRE_DIMENSION( if90id, i_dimid_lat, len = ilat )
          IF(ierr /= nf90_noerr)THEN
             WRITE(*,*) 'generate_mask: Failed to read lat dimension from netcdf file bathy_meter.nc'
             ierr = 1
             RETURN
          END IF
       
          ierr = NF90_INQUIRE_DIMENSION( if90id, i_dimid_lon, len = ilon )
          IF(ierr /= nf90_noerr)THEN
             WRITE(*,*) 'generate_mask: Failed to read lon dimension from netcdf file bathy_meter.nc'
             ierr = 1
             RETURN
          END IF

          jpiglo = ilon
          jpjglo = ilat

       END IF ! jpiglo or jpjglo unset

       ! Find the variable ID
       ierr = NF90_INQ_VARID( if90id, 'Bathymetry', ivarid )
       IF( ierr /= nf90_noerr)THEN
          WRITE (*,*) 'generate_mask: ERROR: netcdf file does not contain ''Bathymetry'' variable'
          ierr = 1
          RETURN
       END IF

       WRITE (*,FMT="('Dimensions of domain are ',I4,' x ',I4)") jpiglo,jpjglo

       ! Allocate masking arrays used in partitioning
       ALLOCATE(bathy(jpiglo,jpjglo), imask(jpiglo,jpjglo), STAT=ierr)
       IF(ierr /= 0)THEN
          WRITE (*,*) 'ERROR: failed to allocate memory for bathy and mask arrays'
          RETURN
       END IF

       ! Read the whole field - start is 1,1 by default
       ierr = NF90_GET_VAR(if90id, ivarid, bathy, count=(/jpiglo,jpjglo/) )
       IF( ierr /= nf90_noerr )THEN
          WRITE (*,*) 'generate_mask: ERROR: failed to read ''Bathymetry'' variable from netcdf file'
          ierr = 1
          RETURN
       END IF

       ! Smooth the bathymetry if required
       IF(ln_sco)THEN
          CALL smooth_bathy(bathy,ierr)
          IF(ierr /= 0)THEN
             WRITE(*,*) 'generate_mask: ERROR: failed to smooth bathy'
             RETURN
          END IF
       END IF

       ! Set imask using the bathymetry (zero on land, 1 otherwise) over 
       ! the global/zoom domain.
       WHERE ( bathy(:,:) <= 1.0E-20 ) 
          imask = LAND
       ELSEWHERE
          imask = 1
       END WHERE

       WRITE (*,FMT="(/'generate_mask: Land/sea mask set from bathymetry OK.')")

    END IF

  END SUBROUTINE generate_mask


  SUBROUTINE smooth_bathy(inbathy, ierr)
    IMPLICIT none
    !!----------------------------------------------------------------------
    !!                      Routine smooth_bathy
    !!   Replicates the smoothing done on the decomposed domain in zgr_sco()
    !!   in domzgr.F90. However, here the domain is NOT decomposed and
    !!   every processor performs an identical smoothing on the whole model
    !!   bathymetry. This is to ensure that the domain decomposition
    !!   is done using a mask that is the same as that which is eventually
    !!   computed after zgr_sco() has been called. (The smoothing process
    !!   below can (erroneously) change whether grid points are wet or dry.)
    !!----------------------------------------------------------------------
    REAL(wp), INTENT(inout), DIMENSION(:,:) :: inbathy ! The bathymetry to 
                                                       ! be smoothed
    INTEGER, INTENT(out)                    :: ierr
    ! Locals
    INTEGER  :: ji, jj, jl
    INTEGER  :: iip1, ijp1, iim1, ijm1   ! temporary integers
    INTEGER  :: x_size, y_size
    REAL(wp) :: zrmax, zri, zrj
    REAL(wp), PARAMETER :: TOL_ZERO = 1.0E-20_wp ! Any value less than 
                                                 ! this assumed zero
    REAL(wp), ALLOCATABLE, DIMENSION(:,:)   :: zenv, ztmp, zmsk, zbot, &
                                               zscosrf, zhbatt
    !
    !!----------------------------------------------------------------------

    ! Do this because we've not decomposed the domain yet and therefore
    ! jpi,jpj,nlc{i,j} etc. are not set.
    x_size = SIZE(inbathy, 1)
    y_size = SIZE(inbathy, 2)

    ALLOCATE(zenv(x_size,y_size), ztmp(x_size,y_size), zmsk(x_size,y_size), &
             zbot(x_size,y_size), zhbatt(x_size, y_size), &
             zscosrf(x_size,y_size), Stat=ierr)
    IF( ierr /= 0 ) THEN
       WRITE(*,*) 'smooth_bathy: ERROR - failed to allocate workspace arrays'
       RETURN
    END IF

    zscosrf(:,:) = 0._wp      ! ocean surface depth (here zero: no under ice-shelf sea)
    zbot(:,:) = inbathy(:,:)  ! ocean bottom depth
    !                         ! set maximum ocean depth
    inbathy(:,:) = MIN( rn_sbot_max, inbathy(:,:) )

    WHERE( inbathy(:,:) > TOL_ZERO ) inbathy(:,:) = &
                                       MAX( rn_sbot_min, inbathy(:,:) )

    ! use r-value to create hybrid coordinates
    zenv(:,:) = MAX( inbathy(:,:), rn_sbot_min )
    ! 
    ! Smooth the bathymetry (if required)
    !
    jl = 0
    zrmax = 1._wp
  
    WRITE(*,"(/'Smoothing bathymetry....'/)")
                                                       ! ================ !
    DO WHILE( jl <= 10000 .AND. zrmax > rn_rmax )      !  Iterative loop  !
       !                                               ! ================ !
       jl = jl + 1
       zrmax = 0._wp
       zmsk(:,:) = 0._wp

       DO jj = 1, y_size
          DO ji = 1, x_size
             iip1 = MIN( ji+1, x_size )      ! force zri = 0 on last line (ji=ncli+1 to jpi)
             ijp1 = MIN( jj+1, y_size )      ! force zrj = 0 on last row  (jj=nclj+1 to jpj)
             zri = ABS( zenv(iip1,jj  ) - zenv(ji,jj) ) / ( zenv(iip1,jj  ) + zenv(ji,jj) )
             zrj = ABS( zenv(ji  ,ijp1) - zenv(ji,jj) ) / ( zenv(ji  ,ijp1) + zenv(ji,jj) )
             zrmax = MAX( zrmax, zri, zrj )
             
             IF( zri > rn_rmax )   zmsk(ji  ,jj  ) = 1._wp
             IF( zri > rn_rmax )   zmsk(iip1,jj  ) = 1._wp
             IF( zrj > rn_rmax )   zmsk(ji  ,jj  ) = 1._wp
             IF( zrj > rn_rmax )   zmsk(ji  ,ijp1) = 1._wp
          END DO
       END DO

       !
       WRITE(*,"('smooth_bathy : iter=',I5,' rmax=',F8.4,' nb of pt= ',I8)") &
                                             jl, zrmax, INT( SUM(zmsk(:,:) ) )
       !

       ! Copy current surface before next smoothing iteration 
       ztmp(:,:) = zenv(:,:)

       DO jj = 1, y_size
          DO ji = 1, x_size
             iip1 = MIN( ji+1, x_size )     ! last  line (ji=nlci)
             ijp1 = MIN( jj+1, y_size )     ! last  raw  (jj=nlcj)
             iim1 = MAX( ji-1,  1  )      ! first line (ji=nlci)
             ijm1 = MAX( jj-1,  1  )      ! first raw  (jj=nlcj)
             IF( zmsk(ji,jj) == 1._wp ) THEN
                ztmp(ji,jj) =   (                                                                                  &
             &    zenv(iim1,ijp1)*zmsk(iim1,ijp1) + zenv(ji,ijp1)*zmsk(ji,ijp1) + zenv(iip1,ijp1)*zmsk(iip1,ijp1)  &
             &  + zenv(iim1,jj  )*zmsk(iim1,jj  ) + zenv(ji,jj  )*    2._wp     + zenv(iip1,jj  )*zmsk(iip1,jj  )  &
             &  + zenv(iim1,ijm1)*zmsk(iim1,ijm1) + zenv(ji,ijm1)*zmsk(ji,ijm1) + zenv(iip1,ijm1)*zmsk(iip1,ijm1)  &
             &                  ) / (                                                                              &
             &                    zmsk(iim1,ijp1) +               zmsk(ji,ijp1) +                 zmsk(iip1,ijp1)  &
             &  +                 zmsk(iim1,jj  ) +                   2._wp     +                 zmsk(iip1,jj  )  &
             &  +                 zmsk(iim1,ijm1) +               zmsk(ji,ijm1) +                 zmsk(iip1,ijm1)  &
             &                      )
             ENDIF
          END DO
       END DO
       !
       DO jj = 1,y_size
          DO ji = 1,x_size
             IF( zmsk(ji,jj) >= 1._wp-TOL_ZERO ) zenv(ji,jj) = MAX( ztmp(ji,jj), inbathy(ji,jj) )
          END DO
       END DO
       !
       !                                                ! ================ !
    END DO                                              !     End loop     !
    !                                                   ! ================ !
    !
    ! envelop bathymetry saved in zhbatt
    zhbatt(:,:) = zenv(:,:) 

    ! Subtract off rn_sbot_min so can check for land using zenv = LAND (0)
    inbathy(:,:) = zenv(:,:) - rn_sbot_min

  END SUBROUTINE smooth_bathy

END MODULE mask_mod
